from typing import List
import pandas as pd
from pandas.api.types import is_datetime64_any_dtype as is_datetime
from pandas.api.types import is_numeric_dtype, is_string_dtype, is_categorical_dtype


def is_k_anonymized(df: pd.DataFrame, k: int = 3, id_cols: List[str] = None):
    """
    Checks if a dataframe satisfies k-anonymity for the given k. If id_cols is given only these columns are checked
    :param df: dataframe to check for k-anonymity
    :param k: the number samples that need to have the same values
    :param id_cols: optional subset of columns in the dataframe that are exclusively checked for k-anonymity
    :return: boolean indicating wether the dataframe satisfies k-anonymity
    """

    for index, row in df.iterrows():
        if id_cols:
            query = ' & '.join([f'{col} == "{row[col]}"' for col in id_cols])
        else:
            query = ' & '.join([f'{col} == "{row[col]}"' for col in df.columns])
        rows = df.query(query)
        if rows.shape[0] < k:
            return False
    return True


def anonymize(df: pd.DataFrame, k: int = 3, id_cols: List[str] = None) -> pd.DataFrame:
    """
    Attempts to generalize the given dataframe to make it k-anonymized

    :param df: dataframe to check
    :param k:
    :param id_cols: optional parameter specifying a subset of columns in the dataframe to generalize
    :return:
    """
    anon_df = df.copy()
    # If id cols are given anonymize those otherwise use all columns
    for col in id_cols if id_cols else df.columns:
        if is_datetime(df[col]):
            anon_df[col] = generalize_datetime_column(df[col])

        elif is_numeric_dtype(df[col]):
            anon_df[col] = generalize_numeric_column(df[col])

        elif is_string_dtype(df[col]) or is_categorical_dtype(df[col]):
            # TODO categorical/string variable handling
            anon_df[col] = df[col]

        else:
            anon_df[col] = anon_df[col]

    if is_k_anonymized(anon_df, k=k):
        return anon_df

    else:
        print("More generalization required")


def generalize_numeric_column(num_col: pd.Series):
    return num_col


def generalize_datetime_column(date_col: pd.Series, level: int = 2):
    col = pd.to_datetime(date_col)

    if level == 2:
        generalized_col = col.apply(lambda x: x.strftime('m-%Y'))
        return pd.to_datetime(generalized_col)

    elif level == 3:
        generalized_col = col.apply(lambda x: x.strftime('%Y'))
        return pd.to_datetime(generalized_col)



import requests
from io import BytesIO
import tarfile
import pika
import json
from typing import Union, List
from tarfile import TarFile
from dotenv import load_dotenv, find_dotenv
import os
import logging
import base64

LOGGER = logging.getLogger(__name__)


class PHTClient:
    """
    Client class for interacting with PHT services

    """

    def __init__(self, api_url: str, api_port: int = 5555, api_token: str = None, ampq_url: str = None,
                 vault_url: str = None, vault_token: str = None):
        """
        Set up connection parameters for the services (train api and rabbit mq)

        :param api_url: endpoint of the central TrainAPI
        :param api_port:
        :param ampq_url: ampq url containing username and password for connecting to rabbitmq
        :param api_token: token to be passed to the api
        :param vault_url: url of the vault api used for storing
        """
        self.api_url = api_url
        self.port = api_port
        self.token = api_token

        self.vault_url = vault_url
        self.api_headers = None
        self.vault_headers = None
        self._create_headers(api_token, vault_token)
        self.rmq_params = None
        if ampq_url:
            self.rmq_params = pika.URLParameters(ampq_url)

    def publish_message_rabbit_mq(self, message: Union[str, bytes, List[str], dict], exchange: str = "pht",
                                  exchange_type: str = "topic", routing_key: str = "pht"):
        """
        Publish a message to rabbit mq with the given message parameters

        :param message: the message to be published
        :param exchange: the identifier of the exchange
        :param exchange_type:
        :param routing_key:
        :return:
        """

        if self.rmq_params:
            connection = pika.BlockingConnection(self.rmq_params)
        else:
            LOGGER.info("No connection to rabbit mq specified, attempting connection on localhost")
            connection = pika.BlockingConnection(pika.ConnectionParameters(host='localhost'))
        channel = connection.channel()
        channel.exchange_declare(exchange=exchange, exchange_type=exchange_type, durable=True)

        json_message = json.dumps(message).encode("utf-8")
        channel.basic_publish(exchange=exchange, routing_key=routing_key, body=json_message)

        LOGGER.info(" [x] Sent %r" % json_message)
        connection.close()

    def get_train_files_archive(self, train_id: str, token: str = None, client_id: str = None):
        """
        Get the tar archive containing files for building a train from the UI api

        :param train_id:
        :return:
        """
        endpoint = f"{train_id}/files/download"
        if not token:
            archive = self._get_tar_archive_from_stream(endpoint)
        else:
            archive = self._get_tar_archive_from_stream(endpoint, token=token, client_id=client_id)

        return archive

    def _get_tar_archive_from_stream(self, endpoint: str, params: dict = None,
                                     external_endpoint: bool = False, token: str = None,
                                     client_id: str = None) -> BytesIO:
        """
        Read a stream of tar data from the given endpoint and return an in memory BytesIO object containing the data

        :param endpoint: address relative to this instances api address
        :param params: dictionary containing additional parameters to be passed to the request
        :param external_endpoint: boolean parameter controlling whether the URL where the request is sent should built using
        the combination of api + endpoint or if the connection should be attempted on the raw endpoint string

        :return:
        """
        if external_endpoint:
            url = endpoint
        else:
            url = self.api_url + endpoint
        headers = self._create_api_headers(api_token=token, client_id=client_id)
        with requests.get(url, params=params, headers=headers, stream=True) as r:
            r.raise_for_status()
            file_obj = BytesIO()
            for chunk in r.iter_content():
                file_obj.write(chunk)
            file_obj.seek(0)

        return file_obj

    def get_user_pk(self, user_id):
        """
        Get the public key associated with the given user_id from vault storage

        :param user_id:
        :return: hex string containing an rsa public key
        """
        url = f"{self.vault_url}v1/user_pks/{user_id}"
        r = requests.get(url, headers=self.vault_headers)
        r.raise_for_status()
        data = r.json()["data"]["data"]
        return data["rsa_public_key"]

    def get_station_pk(self, station_id):
        """
        Get the rsa public of the station specified by station id from vault storage

        :param station_id: identifier of the station in vault
        :return: hex string containing an rsa public key
        """
        url = f"{self.vault_url}v1/station_pks/{station_id}"
        r = requests.get(url, headers=self.vault_headers)
        r.raise_for_status()
        public_key = r.json()["data"]["data"]["rsa_station_public_key"]
        return public_key

    def get_multiple_station_pks(self, station_ids: List) -> dict:
        station_pks = {}
        for id in station_ids:
            station_pks[id] = self.get_station_pk(id)
        return station_pks

    def post_route_to_vault(self, train_id, route, periodic=False):
        route = [str(_) for _ in route]
        vault_url = f"{self.vault_url}v1/kv-pht-routes/data/{train_id}"
        print(route)
        payload = {
            "options": {
                "cas": 0
            },
            "data": {
                "harborProjects": route,
                "repositorySuffix": str(train_id),
                "periodic": periodic
            }
        }
        r = requests.post(vault_url, headers=self.vault_headers, data=json.dumps(payload))
        r.raise_for_status()

    def _create_headers(self, api_token, vault_token):
        self.headers = self._create_api_headers(api_token)
        self.vault_headers = {"X-Vault-Token": vault_token}

    def _create_api_headers(self, api_token: str, client_id: str = "TRAIN_BUILDER"):
        auth_string = f"{client_id}:{api_token}"
        auth_string = base64.b64encode(auth_string.encode("utf-8")).decode()
        headers = {"Authorization": f"Basic {auth_string}"}
        return headers




import json
import pickle
import os

from train_lib.security.HomomorphicAddition import secure_addition
from train_lib.security.KeyManager import KeyManager


class Train:
    def __init__(self,  results=None, query=None):
        """

        :param results:
        :param query:
        """
        self.results = results
        self.query = query
        self.key_manager = KeyManager(train_config='/opt/train_config.json')

    def load_results(self):
        """
        If a result file exists, loads the results. Otherwise will return empty results.
        :return:
        """
        try:
            with open('/opt/pht_results/' + self.results, 'rb') as results_file:
                return pickle.load(file=results_file)
        except Exception:
            return {'analysis': {}, 'discovery': {}}

    def save_results(self, results):
        """
        Saves the result file of the train
        :param results:
        :return:
        """
        try:
            if not os.path.isdir('/opt/pht_results'):
                os.makedirs('/opt/pht_results')
                print('Created results directory')
            with open('/opt/pht_results/' + self.results, 'wb') as results_file:
                return pickle.dump(results, results_file)
        except Exception:
            raise FileNotFoundError("Result file cannot be saved")

    def load_queries(self):
        try:
            with open('/opt/' + self.query, 'r') as query:
                data = json.load(query)

            query_lst = data['query_lst']
            output_list = data['output_param_list']
            media = data["media"]
            return query_lst, output_list, media
        except Exception as e:
            print(e)
            data = {
                        "query_lst": [["link", "NF-CORE-Station1"],
                                      ["link", "NF-CORE-Station2"],
                                      ["link", "NF-CORE-Station3"]],
                        "output_param_list": ["id", "gender", "birthDate"],
                        "media": "False"
                    }
            print("Error in query - use hardcoded query!")
            query_lst = data['query_lst']
            output_list = data['output_param_list']
            media = data["media"]
            return query_lst, output_list, media

    def save_queries(self, query_file):
        """
        :param query:
        :return:
        """
        with open('/opt/' + self.query, 'w') as queries:
            return json.dump(query_file, queries, indent=4)

    def secure_addition_avg(self, total_age, num_pat):
        result = self.load_results()
        try:
            prev_num_pat = result['discovery']['secure_num_pat']
            prev_total_age = result['discovery']['secure_total_age']

            print("Previous secure addition value total number patients:\n"
                  "{}\nand total age\n{}".format(prev_num_pat, prev_total_age))
        except KeyError:
            print("Previous secure addition empty")
            prev_num_pat = None
            prev_total_age = None
        try:
            n = self.key_manager.get_security_param(param="user_he_key")
            print('Using he key from user in train config {}'.format(n))
        except Exception as e:
            print(e)
            print("Errors: no users HE key - use default n")

            n = 26186875800526071848173346729411495257
        result['discovery']['secure_num_pat'] = secure_addition(num_pat, prev_num_pat, int(n))
        result['discovery']['secure_total_age'] = secure_addition(total_age, prev_total_age, int(n))

        return result




from io import BytesIO
from typing import List, Union
import os
import docker
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPublicKey, RSAPrivateKey
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.primitives import serialization
from uuid import uuid4
from loguru import logger
import tarfile


def build_test_train(master_img: str = "harbor-pht.tada5hi.net/master/python/ubuntu",
                     executable: str = "python",
                     train_image_name: str = None,
                     train_files: List[Union[BytesIO, str]] = None,
                     train_dir: Union[str, os.PathLike] = None,
                     entrypoint_file: str = None,
                     private_key: Union[str, os.PathLike, BytesIO, RSAPrivateKey] = None,
                     user_private_key: Union[str, os.PathLike, BytesIO, RSAPrivateKey] = None,
                     security_protocol: bool = False):
    client = docker.from_env()

    if not (train_files or train_dir):
        raise ValueError("Either files or a directory containing train files need to be set.")

    docker_file = _make_docker_file(master_img, executable, entrypoint_file)

    if not train_image_name:
        train_image_name = f"test_train/{uuid4()}"

    logger.info("Building test train image: {}", train_image_name)

    train_img = build_train_image(client=client, docker_file=docker_file, train_files=train_files, train_dir=train_dir,
                                  train_image_name=train_image_name, private_key=private_key,
                                  user_private_key=user_private_key, security_protocol=security_protocol)


def build_train_image(client: docker.DockerClient,
                      docker_file: BytesIO,
                      train_files: List[Union[BytesIO, str]] = None,
                      train_dir: Union[str, os.PathLike] = None,
                      train_image_name: str = None,
                      private_key: Union[str, os.PathLike, BytesIO, RSAPrivateKey] = None,
                      user_private_key: Union[str, os.PathLike, BytesIO, RSAPrivateKey] = None,
                      security_protocol: bool = False
                      ):
    img, build_logs = client.images.build(fileobj=docker_file, tag=train_image_name, rm=True, pull=True)

    container = client.containers.create(img)
    _add_train_files(container, train_files, train_dir)

    if security_protocol:
        config = _make_train_config(private_key, user_private_key)

    container.commit(repository=train_image_name, tag="latest")
    container.wait()
    container.remove()
    logger.info("Built test train image: {}", str(img))
    return img


def _add_train_files(container, train_files: List[Union[BytesIO, str]] = None, train_dir: Union[str, os.PathLike] = None):

    if train_files and train_dir:
        raise ValueError("Only one of train_files or train_dir can be set.")

    if train_dir:
        _add_train_dir_to_img(container, train_dir)
    if train_files:
        _add_list_of_files_to_image(container, train_files)


def _add_train_dir_to_img(container, train_dir: Union[str, os.PathLike] = None):
    train_archive = BytesIO()
    with tarfile.open(fileobj=train_archive, mode="w") as tar:
        tar.add(train_dir, arcname=os.path.basename(train_dir))

    train_archive.seek(0)

    container.put_archive("/opt/pht_train/", train_archive)


def _add_list_of_files_to_image(img, train_files: List[Union[BytesIO, str]] = None):
    raise NotImplementedError()


def _make_train_config(private_key: Union[str, os.PathLike, BytesIO, RSAPrivateKey],
                       user_private_key: Union[str, os.PathLike, BytesIO, RSAPrivateKey]):
    if not private_key:
        # todo generate new private/public key and save/print it
        user_sk, user_pk = _generate_key_pair(output_format="hex")

    if not user_private_key:
        user_sk, user_pk = _generate_key_pair(output_format="hex")


def _generate_key_pair(output_format: str = "pem"):
    private_key = rsa.generate_private_key(public_exponent=65537, key_size=2048)
    public_key = private_key.public_key()

    private_bytes = private_key.private_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PrivateFormat.PKCS8,
        encryption_algorithm=serialization.NoEncryption()
    )
    public_bytes = public_key.public_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PublicFormat.SubjectPublicKeyInfo
    )
    if output_format == "pem":
        return private_bytes, public_bytes

    elif output_format == "hex":

        return private_bytes.hex(), public_bytes.hex()

    return private_key, public_key


def _make_docker_file(master_img: str, executable: str, entrypoint_file: str = None) -> BytesIO:
    if not entrypoint_file:
        entrypoint_file = "/opt/pht_train/entrypoint.py"
    docker_file_obj = BytesIO(
        f"""
        FROM {master_img}
        
        RUN mkdir /opt/pht_results && mkdir /opt/pht_train && chmod -R 755 /opt/pht_train
        CMD ["{executable}", "{entrypoint_file}"]
        """.encode("utf-8")
    )

    return docker_file_obj


if __name__ == '__main__':
    build_test_train(train_dir="../train", entrypoint_file="/opt/pht_train/train/FHIRAverageAgeTrain.py")



import json
import pickle
import os
import scipy.io as sio
import pandas as pd
import shutil
from glob import glob
import matplotlib.pyplot as plt
import numpy as np
from pathlib import Path

COLOR_DICT = {
    'acc': '#e41a1c',
    'sens': '#377eb8',
    'loss': '#ffff33',
    'wacc': '',
    'auc': '#4daf4a',
    'spec': '',
    'f1': '#984ea3',
    'train_loss': '#ff7f00',
}


class Train:
    def __init__(self,  results=None, query=None):
        """

        :param results:
        :param query:
        """
        self.results = results
        self.query = query

    def load_results(self):
        """
        If a result file exists, loads the results. Otherwise will return empty results.
        :return:
        """
        try:
            if not os.path.isdir('/opt/pht_results'):
                os.makedirs('/opt/pht_results')
                print('Created results directory')
            with open('/opt/pht_results/' + self.results, 'rb') as results_file:
                return pickle.load(file=results_file)
        except Exception:
            return {'analysis': {}, 'discovery': {}, 'exec': []}

    def save_results(self, results):
        """
        Saves the result file of the train
        :param results:
        :return:
        """
        try:
            with open('/opt/pht_results/' + self.results, 'wb') as results_file:
                return pickle.dump(results, results_file)
        except Exception:
            raise FileNotFoundError("Result file cannot be saved")

    def load_queries(self):
        """

        :return:
        """
        try:
            with open('/opt/pht_train/' + self.query, 'r') as queries:
                return json.load(queries)
        except Exception:
            return {'1': 'station_1',
                    '2': 'station_2',
                    '3': 'station_3'}

    def save_queries(self, query):
        """

        :param query:
        :return:
        """
        with open('/opt/pht_train/' + self.query, 'w') as queries:
            return json.dump(query, queries)

    def discovery(self, patients, results):
        # discovery
        gender = {'female': 1, 'male': 0, 'unknown': 999}
        patients.gender = [gender[item] for item in patients.gender]
        patients.astype({"gender": 'int'})

        #print(patients.describe())

        discovery = {'total_samples': len(patients),
                     'MEL': patients[patients.note == 'MEL'].shape[0],
                     'NV': patients[patients.note == 'NV'].shape[0],
                     'BCC': patients[patients.note == 'BCC'].shape[0],
                     'AK': patients[patients.note == 'AK'].shape[0],
                     'BKL': patients[patients.note == 'BKL'].shape[0],
                     'DF': patients[patients.note == 'DF'].shape[0],
                     'VASC': patients[patients.note == 'VASC'].shape[0],
                     'SCC': patients[patients.note == 'SCC'].shape[0],
                     'UNK': patients[patients.note == 'UNK'].shape[0]
                     }
        results['discovery']['discovery_exec_' + str(len(results['discovery']) + 1)] = discovery
        print(results)
        return results

    def plot_discovery_results(self, results):
        print(results)
        num_stations = len(results.items()) + 1  # for all samples

        # data to plot
        class_0 = []
        class_1 = []
        class_2 = []
        class_3 = []
        class_4 = []
        class_5 = []
        class_6 = []
        class_7 = []
        class_8 = []

        stat_samples = []

        for i in results.items():
            stat_samples.append(i[1]['total_samples'])
            class_0.append(i[1]['MEL'])
            class_1.append(i[1]['NV'])
            class_2.append(i[1]['BCC'])
            class_3.append(i[1]['AK'])
            class_4.append(i[1]['BKL'])
            class_5.append(i[1]['DF'])
            class_6.append(i[1]['VASC'])
            class_7.append(i[1]['SCC'])
            class_8.append(i[1]['UNK'])

        # append total column
        class_0.append(sum(class_0))
        class_1.append(sum(class_1))
        class_2.append(sum(class_2))
        class_3.append(sum(class_3))
        class_4.append(sum(class_4))
        class_5.append(sum(class_5))
        class_6.append(sum(class_6))
        class_7.append(sum(class_7))
        class_8.append(sum(class_8))

        # create plot
        fig, ax = plt.subplots()
        index = np.arange(num_stations)
        bar_width = 0.5
        opacity = 0.8

        p1 = plt.bar(index, class_0, bar_width, alpha=opacity, label='MEL')
        p2 = plt.bar(index, class_1, bar_width, bottom=class_0, alpha=opacity, label='NV')
        p3 = plt.bar(index, class_2, bar_width, bottom=class_1, alpha=opacity, label='BCC')
        p4 = plt.bar(index, class_3, bar_width, bottom=class_2, alpha=opacity, label='AK')
        p5 = plt.bar(index, class_4, bar_width, bottom=class_3, alpha=opacity, label='BKL')
        p6 = plt.bar(index, class_5, bar_width, bottom=class_4, alpha=opacity, label='DF')
        p7 = plt.bar(index, class_6, bar_width, bottom=class_5, alpha=opacity, label='VASC')
        p8 = plt.bar(index, class_7, bar_width, bottom=class_6, alpha=opacity, label='SCC')
        p9 = plt.bar(index, class_8, bar_width, bottom=class_7, alpha=opacity, label='UNK')

        plt.xlabel('Stations')
        plt.ylabel('Samples')
        plt.title('Samples at different Station')

        station_labels = list(range(1, len(results) + 1))
        station_labels = [str(x) for x in station_labels]
        station_labels.append('All')

        plt.xticks(index, station_labels)
        plt.legend((p1[0], p2[0], p3[0], p4[0], p5[0], p6[0], p7[0], p8[0], p9[0]),
                   ('Class: MEL', 'Class: NV', 'Class: BCC', 'Class: AK', 'Class: BKL', 'Class: DF',
                    'Class: VASC', 'Class: SCC', 'Class: UNK'))

        plt.tight_layout()
        plt.savefig('/opt/pht_results/discovery.png')

    def station_to_dict(self, station_path):
        # TODO: Assumes epochs of first CV! If first CV only one with that amount of epochs ignores complete rest!
        pathlist = glob(station_path + '/**progression_valInd.mat')
        # print(list(pathlist))
        paths = []
        station_dict = {}
        cv_counts = 0
        for path in pathlist:
            mat = sio.loadmat(path)
            cv_counts += 1
            for k, v in mat.items():
                if k in COLOR_DICT.keys():
                    if k not in station_dict.keys():
                        station_dict[k] = v.copy()
                        station_dict[f'{k}_min'] = v.copy()
                        station_dict[f'{k}_max'] = v.copy()
                    else:
                        try:
                            station_dict[k] += v.copy()
                        except Exception as e:
                            print(e)
                            cv_counts -= 1
                            break

                        station_dict[f'{k}_min'] = np.minimum(station_dict[f'{k}_min'], v.copy())
                        station_dict[f'{k}_max'] = np.maximum(station_dict[f'{k}_max'], v.copy())

                else:
                    if not k in station_dict.keys():
                        station_dict[k] = v
        for k in station_dict.keys():
            if k in COLOR_DICT.keys():
                station_dict[k] /= cv_counts

        _st = Path(pathlist[0]).parents[1].name
        station_dict['station'] = _st[0].upper() + _st[1:]

        return station_dict

    def make_plot_dict(self, station_path, plot_params, std=False, num_stations=3):
        '''
        This helper function encapsulates the generation of a plot dictionary containing all information relevant for the
        plot and improving the possibility to

        Args:
            station_path: Path or list of paths to station results
            plot_params: parameters which to be plotted given as list
            std: Add range of all CVs to the plot
            num_stations: Number of stations in the end (extends plot)

        Returns: plot dictionary for a given route

        '''

        if not isinstance(station_path, list):
            station_path = [station_path]

        plot_dict = {
            'num_stations': len(station_path),
            'stations': [],
            'station_epochs': [],
            'params': {},
            'plot_epochs': np.array([])
        }
        for p in plot_params:
            plot_dict['params'][p] = np.array([])
            if std:
                plot_dict['params'][f'{p}_min'] = np.array([])
                plot_dict['params'][f'{p}_max'] = np.array([])

        for station in station_path:
            m = self.station_to_dict(station)
            plot_dict['station_epochs'].append(m['step_num'][0][-1])
            plot_dict['stations'].append(m['station'])
            if not plot_dict['plot_epochs'].size == 0:
                plot_dict['plot_epochs'] = np.append(plot_dict['plot_epochs'], plot_dict['plot_epochs'][-1] + m['step_num'][0])
            else:
                plot_dict['plot_epochs'] = np.append(plot_dict['plot_epochs'], m['step_num'][0])
            for k in plot_dict['params'].keys():
                if k in ['acc', 'f1', 'train_loss', 'loss']:
                    plot_dict['params'][k] = np.append(plot_dict['params'][k], m[k][0])
                    if std:
                        plot_dict['params'][f'{k}_min'] = np.append(plot_dict['params'][f'{k}_min'], m[f'{k}_min'][0])
                        plot_dict['params'][f'{k}_max'] = np.append(plot_dict['params'][f'{k}_max'], m[f'{k}_max'][0])
                elif k in ['sens', 'auc']:
                    plot_dict['params'][k] = np.append(plot_dict['params'][k], np.mean(m[k], axis=1))
        return plot_dict

    def central_plot(self, routes, plot_params, save_path, smoothing_factor=None, std=False, save=False, num_stations=1):
        smoothing = False
        if smoothing_factor:
            smoothing_factor = np.ones(len(plot_params)) * smoothing_factor if isinstance(smoothing_factor,float) \
                else np.array(smoothing_factor)
            smoothing = True
        else:
            smoothing_factor = np.ones(len(plot_params))

        nrows = 1
        ncols = 1
        fig, axs = plt.subplots(nrows=nrows, ncols=ncols, sharex='row', sharey='all', squeeze=False, figsize=(6, 5))

        for i, ax in enumerate(fig.axes):
            plot_dict = self.make_plot_dict(routes[i], plot_params, std=std)
            extend_plot = num_stations - plot_dict['num_stations'] if num_stations >= plot_dict['num_stations'] else plot_dict['num_stations']

            ax.set_xlim(right=(plot_dict['num_stations']+extend_plot)*60)
            ax.set_ylim(bottom=0.2, top=0.9)

            min_val = np.inf
            max_val = -np.inf
            plotting = {item: value for (item, value) in plot_dict['params'].items() if item in plot_params}
            # for (k, v), s in zip(plot_dict['params'].items(), smoothing_factor):
            for (k, v), s in zip(plotting.items(), smoothing_factor):
                min_val = np.min(v) if np.min(v) < min_val else min_val
                max_val = np.max(v) if np.max(v) > max_val else max_val

                if k in ['acc', 'sens', 'f1']:
                    measure = np.max
                else:
                    measure = np.min

                print(f'Measure "{k}": best overall value -> {measure(v)}')
                print(f'Measure "{k}": last value -> {v[-1]}')

                ax.plot(plot_dict['plot_epochs'], v,
                        color=COLOR_DICT[k],
                        label=k)
                if smoothing:
                    ax.plot(plot_dict['plot_epochs'], np.array(pd.Series(v).ewm(alpha=s).mean().to_numpy()),
                            color=COLOR_DICT[k],
                            label=f'{k}_smoothed',
                            alpha=.7)

                if std:
                    ax.fill_between(plot_dict['plot_epochs'], plot_dict['params'][f'{k}_max'], plot_dict['params'][f'{k}_min'],
                                    alpha=.2, color=COLOR_DICT[k])

            div = 0
            for station in plot_dict['station_epochs'][:-1]:
                div += station
                ax.axvline(div, linestyle='--')

            ax.legend(loc='lower right', framealpha=0.5, title='Metrics')
            if i%ncols == 0:
                ax.set_ylabel('Result')

            ax.set_xlabel('Epochs')
            ax.set_title('Evaluation of Efficientnet-b6 on ISIC 2019')
            handles, labels = ax.get_legend_handles_labels()

        plt.subplots_adjust(left=0.15, bottom=0.2, right=0.95)

        if save:
            plt.savefig(save_path)
        else:
            plt.show()

    def plot_stations(self, routes, plot_params, save_path, smoothing_factor=None, std=False, save=False, num_stations=3):
        smoothing = False
        if smoothing_factor:
            smoothing_factor = np.ones(len(plot_params)) * smoothing_factor if isinstance(smoothing_factor,float) \
                else np.array(smoothing_factor)
            smoothing = True
        else:
            smoothing_factor = np.ones(len(plot_params))

        nrows = 1
        ncols = 1
        fig, axs = plt.subplots(nrows=nrows, ncols=ncols, sharex='row', sharey='all', squeeze=True, figsize=(6, 5))

        for i, ax in enumerate(fig.axes):
            plot_dict = self.make_plot_dict(routes[i], plot_params, std=std)
            extend_plot = num_stations - plot_dict['num_stations'] if num_stations >= plot_dict['num_stations'] else plot_dict['num_stations']

            ax.set_xlim(right=(plot_dict['num_stations']+extend_plot)*20)
            ax.set_ylim(bottom=0.2, top=.9)

            min_val = np.inf
            max_val = -np.inf
            plotting = {item: value for (item, value) in plot_dict['params'].items() if item in plot_params}
            # for (k, v), s in zip(plot_dict['params'].items(), smoothing_factor):
            for (k, v), s in zip(plotting.items(), smoothing_factor):
                min_val = np.min(v) if np.min(v) < min_val else min_val
                max_val = np.max(v) if np.max(v) > max_val else max_val

                if k in ['acc', 'sens', 'f1']:
                    measure = np.max
                else:
                    measure = np.min

                print(f'Measure "{k}": best overall value -> {measure(v)}')
                print(f'Measure "{k}": last value -> {v[-1]}')

                ax.plot(plot_dict['plot_epochs'], v,
                        color=COLOR_DICT[k],
                        label=k)
                if smoothing:
                    ax.plot(plot_dict['plot_epochs'], np.array(pd.Series(v).ewm(alpha=s).mean().to_numpy()),
                            color=COLOR_DICT[k],
                            label=f'{k}_smoothed',
                            alpha=.7)

                if std:
                    ax.fill_between(plot_dict['plot_epochs'], plot_dict['params'][f'{k}_max'], plot_dict['params'][f'{k}_min'],
                                    alpha=.2, color=COLOR_DICT[k])

            div = 0
            for station in plot_dict['station_epochs'][:-1]:
                div += station
                ax.axvline(div, linestyle='--')

            ax.legend(loc='lower right', framealpha=0.5, title='Metrics')
            if i%ncols == 0:
                ax.set_ylabel('Result')
            ax.set_xlabel('Epochs')
            for j, st in enumerate(['Station 1', 'Station 2', 'Station 3']):
                ax.text(6+(20*j), 0.95, st, transform=ax.get_xaxis_transform())

            ax.set_title('Evaluation of Efficientnet-b6 on ISIC 2019')
            handles, labels = ax.get_legend_handles_labels()

        plt.subplots_adjust(left=0.15, bottom=0.2, right=0.95)

        if save:
            plt.savefig(save_path)
        else:
            plt.show()

    def create_label_file(self, patients_df, path_labels, station):
        patients_df.patientId = patients_df.patientId.str.replace('-', '_')
        y = pd.get_dummies(patients_df.note)
        cols = ["MEL", "NV", "BCC", "AK", "BKL", "DF", "VASC", "SCC"]
        new_index_df = y[cols].astype('float')
        new_index_df["UNK"] = 0.0
        new_index_df.insert(0, "image", patients_df.patientId)
        print(new_index_df.describe())
        new_index_df.to_csv(path_labels + 's_' + str(station) + '_labels.csv', index=False)

        return new_index_df

    def move_file(self, src_file, dst_file):
        shutil.move(src=src_file, dst=dst_file)
        return True

    def copy_file(self, src_file, dst_file):
        shutil.copy(src=src_file, dst=dst_file)
        return True

    def create_stat_res_dirs(self, station):
        os.mkdir('/opt/pht_results/res_' + str(station))
        return True

    def locate_mat_files(self, exec_path):
        return glob(exec_path + '/**/*.mat')

    def find_best_cv(self, performance):
        res = {}
        max_val = 0
        for key, nested in performance.items():
            if nested['BestF1'] > max_val:
                max_val = nested['BestF1']
                res[key] = max_val
        print('Best CV performance in fold {} (index 0-4)'.format(list(res)[-1]))
        return list(res)[-1]

    def clean_up(self, exec_dir):
        os.rmdir(exec_dir)
        return True




import pandas as pd
import os
import json
import datetime


RESULTS_PATH = "/opt/pht_results/average_age.json"


def load_previous_data(path):
    if os.path.exists(path):
        with open(path, "r") as f:
            average_age_dict = json.load(f)

        return average_age_dict

    else:
        return None


def age_from_dob(dob):
    today = datetime.date.today()
    return today.year - dob.year - ((today.month, today.day) < (dob.month, dob.day))


def calculate_new_average(average_age_dict, data_path, results_path):
    # load the data and ensure that birthdate is a datetime column
    data = pd.read_csv(data_path)
    data["birthDate"] = pd.to_datetime(data["birthDate"])

    ages = data["birthDate"].apply(lambda x: age_from_dob(x))

    local_average = ages.mean()

    # previous results exist load them otherwise create a new dictionary containing the results
    if average_age_dict:
        prev_average = average_age_dict["average_age"]
        new_average = (prev_average + local_average) / 2 if prev_average else local_average
        average_age_dict["average_age"] = new_average
    else:
        new_average = local_average
        average_age_dict = {"average_age": new_average}

    print(average_age_dict)

    # store the updated results
    with open(results_path, "w") as f:
        json.dump(average_age_dict, fp=f, indent=2)


def main():
    data_path = os.getenv("TRAIN_DATA_PATH", "/opt/train_data/patients.csv")
    print(f"Loading data at {data_path}")
    prev_results = load_previous_data(RESULTS_PATH)
    calculate_new_average(prev_results, data_path, RESULTS_PATH)


if __name__ == '__main__':
    main()



import subprocess
from typing import List, Tuple


def validate_train_image(train_img: str, master_image: str):
    """
    Validates a train image against an official master image
    :param train_img: identifier of the docker image defining a train
    :param master_image: identifier of the master docker image to validate against
    :return:
    """

    status, message = _compare_image_file_system(master_image, train_img)
    print(message)
    if status != 0:
        raise ValueError(f"File system could not be validated. \n {message}")


def _compare_image_file_system(master_image_name: str, train_image_name: str):
    """
    Compares the full file systems of the master image against, any added, changed or deleted files will be detected
    and if any changes are detected outside of the PHT specific directories will be considered invalid and raise an
    error.
    Uses the container-diff tool from google: https://github.com/GoogleContainerTools/container-diff

    :param master_image_name:
    :param train_image_name:
    :return:
    """

    container_diff_args = ["container-diff", "diff", f"daemon://{master_image_name}",
                           f"daemon://{train_image_name}", "--type=file"]
    output = subprocess.run(container_diff_args, capture_output=True)
    file_system_diff = output.stdout.decode().splitlines()
    valid, msg = _validate_file_system_changes(file_system_diff)
    if valid:
        return 0, "No file system anomalies detected"
    else:
        return 1, "Invalid file system changes detected, files can only be added into " \
                  f"/opt/pht_train, but found {msg}"


def _validate_file_system_changes(file_system_diff: List[str]) -> Tuple[bool, str]:
    """
    Validate the file system changes found by the container-diff tools.
    Checks if files have been added at the right location and whether files have been deleted or changed compared
    to the master image

    :param file_system_diff: output generated by the container-diff tool analysing the file system changes between
    two images

    :return: whether the detected changes to the file system are valid
    """
    add_ind = None
    deleted_ind = None
    changed_ind = None
    valid = True
    for ind, content in enumerate(file_system_diff):
        if "These entries have been added" in content:
            add_ind = ind
        elif "These entries have been deleted" in content:
            deleted_ind = ind
        elif "These entries have been changed" in content:
            changed_ind = ind
    # Find the files added to the image file system and make sure they are located exclusively under /opt/pht_train
    if len(file_system_diff[add_ind: deleted_ind]) > 2:
        print("Added files detected.")
        valid = True
        invalid_files = []
        for file in file_system_diff[add_ind + 2: deleted_ind]:
            valid_file, file = _validate_added_file(file)
            if not valid_file:
                valid = False
                invalid_files.append(file)
        invalid_file_string = "\n".join(invalid_files)
        if not valid:
            return False, f"Incorrectly added files:\n{invalid_file_string} "
    # If the length of the deleted files section is greater than two, files have been deleted from the master image
    # -> image invalid
    if len(file_system_diff[deleted_ind: changed_ind]) > 2:
        print("Deleted Files detected")
        print(file_system_diff[deleted_ind: changed_ind])
        valid = False
        return valid, "Files deleted from master image"
    # If the length of the deleted files section is greater than two, files have been changed from the master image
    # -> image invalid
    if len(file_system_diff[changed_ind:]) > 2:
        print("Changed files detected")
        valid = False
        return valid, "Files changed in the master"

    return valid, "Successfully verified file system"


def _validate_added_file(file: str) -> Tuple[bool, str]:
    """
    Checks whether an added file detected by container-diff is located under /opt/pht_train.

    :param file: line of output generated by container diff containing info on the added file
    :return: whether the file is correctly located or not
    """
    path = file.split(" ")[0]
    valid = False

    print(f"Validate called with file: {file}")
    if not file:
        return True, path

    if len(path) > 1:
        path_dir = path.split("/")[1:]

        if path_dir[0] == "opt":

            if path_dir[1] == "pht_results":
                valid = True
            if path_dir[1] == "pht_train":
                valid = True

            if path_dir[1] == "train_config.json":
                valid = True

            if path_dir[1] == "user_sym_key.key":
                valid = True
    if not valid:
        print(f"Invalid file detected: {path}")

    return valid, path



import json
import pickle
import os

from train_lib.security.HomomorphicAddition import secure_addition
from train_lib.security.KeyManager import KeyManager


class Train:
    def __init__(self,  results=None, query=None):
        """

        :param results:
        :param query:
        """
        self.results = results
        self.query = query
        self.key_manager = KeyManager(train_config='/opt/train_config.json')

    def load_results(self):
        """
        If a result file exists, loads the results. Otherwise will return empty results.
        :return:
        """
        try:
            if not os.path.isdir('/opt/pht_results'):
                os.makedirs('/opt/pht_results')
                print('Created results directory')
            with open('/opt/pht_results/' + self.results, 'rb') as results_file:
                return pickle.load(file=results_file)
        except Exception:
            return {'analysis': {}, 'discovery': {}, 'exec': []}

    def save_results(self, results):
        """
        Saves the result file of the train
        :param results:
        :return:
        """
        try:
            with open('/opt/pht_results/' + self.results, 'wb') as results_file:
                return pickle.dump(results, results_file)
        except Exception:
            raise FileNotFoundError("Result file cannot be saved")

    def load_queries(self):
        """

        :return:
        """
        try:
            with open('/opt/pht_train/' + self.query, 'r') as queries:
                return json.load(queries)
        except Exception:
            return {'1': 'Station1',
                    '2': 'Station2',
                    '3': 'Station3'}

    def save_queries(self, query):
        """

        :param query:
        :return:
        """
        with open('/opt/pht_train/' + self.query, 'w') as queries:
            return json.dump(query, queries)

    def get_user_pk(self):
        try:
            with open('/opt/train_config.json', 'r') as train_conf:
                conf = json.load(train_conf)
                return conf['user_secure_add_pk']
        except Exception:
            return {'user_secure_add_pk': None}

    def secure_addition(self, local_result):
        result = self.load_results()
        try:
            prev_result = result['analysis']['task_a']
            print("Previous secure addition value {}".format(prev_result))
        except KeyError:
            print("Previous secure addition empty")
            prev_result = None
        try:
            n = self.key_manager.get_security_param(param="user_he_key")
        except Exception as e:
            print("Cannot load users he_key - use default n")
            print(e)
            n = 261846875800526071848173346729411495257

        return secure_addition(local_result, prev_result, int(n))


from io import BytesIO

from cryptography.fernet import Fernet
from typing import List, Union, BinaryIO
import logging


class FileEncryptor:
    """
    Performs symmetric encryption and decryption of sensitive files belonging to the train cargo
    """

    def __init__(self, key: bytes):
        self.fernet = Fernet(key)

    def encrypt_files(self, files: Union[List[str], List[BinaryIO]], binary_files=False) -> Union[List[BytesIO], None]:
        """
        Decrypt the given files using symmetric encryption
        :return:
        """
        logging.info("Encrypting files..")
        if binary_files:
            encr_files = []
            for i, file in enumerate(files):
                logging.info(f"file {i + 1}/{len(files)}...")
                # Encrypt the files and convert them to bytes io file objects
                data = file.read()
                encr_files.append(BytesIO(self.fernet.encrypt(data)))
                logging.info("Done")
            return encr_files

        for i, file in enumerate(files):
            logging.info(f"File {i + 1}/{len(files)}...")
            with open(file, "rb") as f:
                encr_file = self.fernet.encrypt(f.read())
            with open(file, "wb") as ef:
                ef.write(encr_file)
            logging.info("Done")

    def decrypt_files(self, files: Union[List[str], List[BinaryIO]], binary_files=False) -> Union[List[BytesIO], None]:
        """
        Decrypt the given files using symmetric encryption
        :return:
        """
        logging.info("Decrypting files..")
        if binary_files:
            decr_files = []
            for i, file in enumerate(files):
                logging.info(f"file {i + 1}/{len(files)}...")
                data = self.fernet.decrypt(file.read())
                decr_files.append(BytesIO(data))
                logging.info("Done")
            return decr_files
        for i, file in enumerate(files):
            logging.info(f"File {i + 1}/{len(files)}...")
            with open(file, "rb") as f:
                decr_file = self.fernet.decrypt(f.read())
            with open(file, "wb") as ef:
                ef.write(decr_file)
            logging.info("Done")



import pickle
from sklearn.naive_bayes import MultinomialNB
from sklearn.model_selection import train_test_split
import os
import numpy as np
import matplotlib.pyplot as plt
import json

from train_lib.security.HomomorphicAddition import secure_addition
from train_lib.security.KeyManager import KeyManager

class Train:
    def __init__(self, model=None, results=None, query=None):
        # Model and results encoded with Pickle
        self.encoded_model = model
        self.results = results
        self.query = query

    def _load_model(self):
        with open(self.encoded_model, 'rb') as model_file:
            return pickle.load(file=model_file)

    def _save_model(self, model):
        with open(self.encoded_model, 'wb') as model_file:
            pickle.dump(model, model_file)

    def _load_results(self):
        try:
            with open('/opt/pht_results/' + self.results, 'rb') as results_file:
                return pickle.load(file=results_file)
        except:
            return {'analysis': {}, 'discovery': {}}

    def _save_results(self, results):
        try:
            if not os.path.isdir('/opt/pht_results'):
                os.makedirs('/opt/pht_results')
                print('Created results directory')
            with open('/opt/pht_results/' + self.results, 'wb') as results_file:
                return pickle.dump(results, results_file)
        except Exception:
            raise FileNotFoundError("Result file cannot be saved")

    def load_queries(self):
        try:
            with open('/opt/' + self.query, 'r') as query:
                data = json.load(query)

            query_lst = data['query_lst']
            output_list = data['output_param_list']
            media = data["media"]
            return query_lst, output_list, media
        except Exception as e:
            print(e)
            data = {
                        "query_lst": [["link", "HIVdemo_LMU,HIVdemo_TUM,HIVdemo_UKT", "birthdate", "le1970-01-01"],
                                        ["link", "HIVdemo_LMU,HIVdemo_TUM,HIVdemo_UKT", "birthdate", "gt1970-01-01", "birthdate", "le2000-01-01"],
                                        ["link", "HIVdemo_LMU,HIVdemo_TUM,HIVdemo_UKT", "birthdate", "gt2000-01-01"]],
                        "output_param_list": ["id", "gender", "birthDate", "link", "observedAllele", "observedSeq"],
                        "media": "MolSeq"
                    }
            print("Error in query - use hardcoded query!")
            query_lst = data['query_lst']
            output_list = data['output_param_list']
            media = data["media"]
            return query_lst, output_list, media

    def save_queries(self, query_file):
        """
        :param query:
        :return:
        """
        with open('/opt/' + self.query, 'w') as queries:
            return json.dump(query_file, queries, indent=4)

    def _flatten(self, l):
        return [item for sublist in l for item in sublist]

    def load_sequences(self, patients, class_labels):
        _RESIDUES = [
            'A', 'R', 'N', 'D', 'C', 'E', 'Q', 'G', 'H', 'I', 'L', 'K', 'M', 'F', 'P', 'S', 'T', 'W', 'Y', 'V', '-'
        ]

        _RESIDUE_ENCODING = {
            residue: (index * [0]) + [1] + (len(_RESIDUES) - index - 1) * [0] for index, residue in enumerate(_RESIDUES)
        }

        x = []
        y = []

        feature_space_dimension = None

        for index, row in patients.iterrows():
            label = 'CXCR4' if len(row["observedAllele"]) > 4 else row["observedAllele"]
            seq = row["observedSeq"]
            seq_encoded = self._flatten(_RESIDUE_ENCODING[residue] for residue in seq)
            seq_encoded_len = len(seq_encoded)

            if feature_space_dimension is None:
                feature_space_dimension = seq_encoded_len
            elif feature_space_dimension != seq_encoded_len:
                raise ValueError('Inconsistent feature length: {} vs. {}'.format(
                    feature_space_dimension, seq_encoded_len))
            x.append(seq_encoded)
            y.append(class_labels[label])
        return x, y

    def plot_results(self, results):
        # figure 1 sample sizes and test size
        num_stations = len(results.items()) + 1  # for all samples

        # data to plot
        train_sam = []
        test_sam = []
        acc = []

        for i in results.items():
            test_sam.append(i[1]['test_samples'])
            train_sam.append(i[1]['training_samples'])
            acc.append(i[1]['acc_total'])

        # append total column
        train_sam.append(sum(train_sam))
        test_sam.append(sum(test_sam))

        # create plot
        plt.figure(1)
        index = np.arange(num_stations)
        bar_width = 0.5
        opacity = 0.8

        p1 = plt.bar(index, train_sam, bar_width, alpha=opacity, color='deepskyblue', label='train')
        p2 = plt.bar(index, test_sam, bar_width, bottom=train_sam, alpha=opacity, color='coral', label='test')

        plt.xlabel('Stations')
        plt.ylabel('Samples')
        plt.title('Samples per Station')

        station_labels = list(range(1, len(results) + 1))
        station_labels = [str(x) for x in station_labels]
        station_labels.append('All')

        # print(p1, p2)
        # print(p1[0], p2[0])

        plt.xticks(index, station_labels)
        plt.legend((p1[0], p2[0]), ('train', 'test'))
        plt.legend()

        plt.tight_layout()

        if not os.path.isdir("/opt/pht_results/"):
            os.mkdir("/opt/pht_results/")

        plt.savefig('/opt/pht_results/analysis_1.png')
        plt.show()

        # figure 2 acc over stations
        plt.figure(2)
        index = np.arange(num_stations)
        p3 = plt.plot(acc, 'ro')

        plt.xlabel('Stations')
        plt.ylabel('Accuracy [%]')
        plt.title('Accuracy over stations')

        station_labels = list(range(1, len(results) + 1))
        station_labels = [str(x) for x in station_labels]

        plt.xticks(np.arange(num_stations - 1), station_labels)
        plt.legend('Accuracy')
        plt.legend()

        plt.tight_layout()
        plt.savefig('/opt/pht_results/analysis_2.png')
        plt.show()



from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

with open("requirements.txt", "r") as rf:
    requirements = rf.readlines()

with open("requirements_dev.txt", "r") as rfd:
    dev_requirements = rfd.readlines()

setup(
    name="pht-train-container-library",
    version="1.0.0",
    author="Michael Graf",
    author_email="michael.graf@uni-tuebingen.de",
    description="PHT train container library",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://gitlab.com/PersonalHealthTrain/implementations/germanmii/difuture/train-container-library.git",
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    keywords=["PHT", "security", "encryption", "personalhealthtrain", "docker"],
    python_requires=">=3.7",
    install_requires=[
        "aiohttp==3.8.0; python_version >= '3.6'",
        "aiosignal==1.2.0; python_version >= '3.6'",
        "async-timeout==4.0.0; python_version >= '3.6'",
        "attrs==21.2.0; python_version >= '2.7' and python_version not in '3.0, 3.1, 3.2, 3.3, 3.4'",
        "certifi==2021.10.8",
        "cffi==1.15.0",
        "charset-normalizer==2.0.7; python_version >= '3'",
        "click>=7.0.0; python_version >= '3.6'",
        "click-spinner==0.1.10",
        "colorama==0.4.4; sys_platform == 'win32' and platform_system == 'Windows' and platform_system == 'Windows'",
        "cryptography==35.0.0",
        "dnspython==2.1.0; python_version >= '3.6'",
        "docker==5.0.3",
        "email-validator==1.1.3",
        "fhir-kindling==0.3.0",
        "fhir.resources==6.1.0; python_version >= '3.6'",
        "fhirpy==1.2.1",
        "frozenlist==1.2.0; python_version >= '3.6'",
        "hvac==0.11.2",
        "idna==3.3; python_version >= '3'",
        "loguru==0.5.3",
        "multidict==5.2.0; python_version >= '3.6'",
        "numpy==1.21.3",
        "oauthlib==3.1.1",
        "orjson==3.6.4; python_version >= '3.7'",
        "pandas==1.3.4",
        "pendulum==2.1.2",
        "pika==1.2.0",
        "pycparser==2.20; python_version >= '2.7' and python_version not in '3.0, 3.1, 3.2, 3.3'",
        "pydantic[email]==1.8.2; python_full_version >= '3.6.1'",
        "pypiwin32==223; sys_platform == 'win32'",
        "python-dateutil==2.8.2; python_version >= '2.7' and python_version not in '3.0, 3.1, 3.2, 3.3'",
        "python-dotenv==0.19.1",
        "pytz==2021.3",
        "pytzdata==2020.1; python_version >= '2.7' and python_version not in '3.0, 3.1, 3.2, 3.3'",
        "pywin32==227; sys_platform == 'win32'",
        "pyyaml==6.0; python_version >= '3.6'",
        "redis==3.5.3",
        "requests==2.26.0",
        "requests-oauthlib==1.3.0",
        "six==1.16.0; python_version >= '2.7' and python_version not in '3.0, 3.1, 3.2, 3.3'",
        "tqdm==4.62.3; python_version >= '2.7' and python_version not in '3.0, 3.1, 3.2, 3.3'",
        "typing-extensions==3.10.0.2; python_version < '3.10'",
        "urllib3==1.26.7; python_version >= '2.7' and python_version not in '3.0, 3.1, 3.2, 3.3, 3.4' and python_version < '4'",
        "websocket-client==1.2.1; python_version >= '3.6'",
        "win32-setctime==1.0.3; sys_platform == 'win32'",
        "xmltodict==0.12.0",
        "yarl==1.7.2; python_version >= '3.6'",
    ],
    tests_require=dev_requirements,
    extras_require={
        "dev": [
            "appdirs==1.4.4",
            "atomicwrites==1.4.0; sys_platform == 'win32'",
            "attrs==21.2.0; python_version >= '2.7' and python_version not in '3.0, 3.1, 3.2, 3.3, 3.4'",
            "backports.entry-points-selectable==1.1.0; python_version >= '2.7'",
            "black==19.10b0; python_version >= '3.6'",
            "bleach==4.1.0; python_version >= '3.6'",
            "bump2version==1.0.1",
            "cached-property==1.5.2",
            "cerberus==1.3.4",
            "certifi==2021.10.8",
            "charset-normalizer==2.0.7; python_version >= '3'",
            "click==8.0.3; python_version >= '3.6'",
            "colorama==0.4.4; sys_platform == 'win32'",
            "coverage==6.0.2",
            "distlib==0.3.3",
            "docutils==0.18; python_version >= '2.7' and python_version not in '3.0, 3.1, 3.2, 3.3, 3.4'",
            "filelock==3.3.1; python_version >= '3.6'",
            "ghp-import==2.0.2",
            "idna==3.3; python_version >= '3'",
            "importlib-metadata==4.8.1; python_version >= '3.6'",
            "iniconfig==1.1.1",
            "jinja2==3.0.2; python_version >= '3.6'",
            "keyring==23.2.1; python_version >= '3.6'",
            "markdown==3.3.4; python_version >= '3.6'",
            "markupsafe==2.0.1; python_version >= '3.6'",
            "mergedeep==1.3.4; python_version >= '3.6'",
            "mkdocs==1.2.3",
            "mkdocs-material==7.3.4",
            "mkdocs-material-extensions==1.0.3; python_version >= '3.6'",
            "orderedmultidict==1.0.1",
            "packaging==20.9; python_version >= '2.7' and python_version not in '3.0, 3.1, 3.2, 3.3'",
            "pathspec==0.9.0",
            "pep517==0.12.0",
            "pip-shims==0.5.3; python_version >= '2.7' and python_version not in '3.0, 3.1, 3.2, 3.3, 3.4'",
            "pipenv-setup==3.1.1",
            "pipfile==0.0.2",
            "pkginfo==1.7.1",
            "platformdirs==2.4.0; python_version >= '3.6'",
            "plette[validation]==0.2.3; python_version >= '2.6' and python_version not in '3.0, 3.1, 3.2, 3.3'",
            "pluggy==1.0.0; python_version >= '3.6'",
            "py==1.10.0; python_version >= '2.7' and python_version not in '3.0, 3.1, 3.2, 3.3'",
            "pygments==2.10.0; python_version >= '3.5'",
            "pymdown-extensions==9.0; python_version >= '3.6'",
            "pyparsing==3.0.2; python_version >= '3.6'",
            "pytest==6.2.5",
            "python-dateutil==2.8.2; python_version >= '2.7' and python_version not in '3.0, 3.1, 3.2, 3.3'",
            "pywin32-ctypes==0.2.0; sys_platform == 'win32'",
            "pyyaml==6.0; python_version >= '3.6'",
            "pyyaml-env-tag==0.1; python_version >= '3.6'",
            "readme-renderer==30.0",
            "regex==2021.10.23",
            "requests==2.26.0",
            "requests-toolbelt==1.0.0",
            "requirementslib==1.5.16; python_version >= '2.7' and python_version not in '3.0, 3.1, 3.2, 3.3, 3.4'",
            "rfc3986==1.5.0",
            "six==1.16.0; python_version >= '2.7' and python_version not in '3.0, 3.1, 3.2, 3.3'",
            "toml==0.10.2; python_version >= '2.6' and python_version not in '3.0, 3.1, 3.2, 3.3'",
            "tomli==1.2.2; python_version >= '3.6'",
            "tomlkit==0.7.2; python_version >= '2.7' and python_version not in '3.0, 3.1, 3.2, 3.3, 3.4'",
            "tox==3.24.4",
            "tqdm==4.62.3; python_version >= '2.7' and python_version not in '3.0, 3.1, 3.2, 3.3'",
            "twine==3.4.2",
            "typed-ast==1.4.3",
            "urllib3==1.26.7; python_version >= '2.7' and python_version not in '3.0, 3.1, 3.2, 3.3, 3.4' and python_version < '4'",
            "virtualenv==20.9.0; python_version >= '2.7' and python_version not in '3.0, 3.1, 3.2, 3.3, 3.4'",
            "vistir==0.5.2; python_version >= '2.7' and python_version not in '3.0, 3.1, 3.2, 3.3'",
            "watchdog==2.1.6; python_version >= '3.6'",
            "webencodings==0.5.1",
            "wheel==0.37.0; python_version >= '2.7' and python_version not in '3.0, 3.1, 3.2, 3.3, 3.4'",
            "zipp==3.6.0; python_version >= '3.6'",
        ]
    },
)



from .KeyManager import KeyManager
from .SymmetricEncryption import FileEncryptor
from .SecurityErrors import ValidationError
from .Hashing import *
from train_lib.docker_util.docker_ops import *
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey
from cryptography.hazmat.primitives.asymmetric import utils, padding
import os
from typing import Union
import time
from tarfile import TarInfo
import docker
import logging


class SecurityProtocol:
    """
    Class that performs the security protocol outlined in the security concept

    :param station_id: PID used to identify the station and to access the correct security values inside the
        train_config.json

    :type station_id: str
    :param config: either a string containing a path to the train_config.json or a dictionary containing the values
        parsed from said json file
    :type config: Union[str, dict]
    :param results_dir: path to the directory containing the results
    :param train_dir: path to the directory containing the immutable files defining a train
    """

    def __init__(self, station_id: str, config: Union[str, dict], results_dir: str = None, train_dir: str = None,
                 docker_client=None):
        self.station_id = station_id
        self.key_manager = KeyManager(train_config=config)
        self.results_dir = results_dir
        self.train_dir = train_dir
        self.docker_client = docker_client
        # self.redis = redis.Redis(decode_responses=True)

    def pre_run_protocol(self, img: str = None, private_key_path: str = None, immutable_dir: str = "/opt/pht_train",
                         mutable_dir: str = "/opt/pht_results"):
        """
        Decrypts the files contained in the train. And performs the steps necessary to validate a train before it is
        being run

        :param img: identifier of the image from which the security relevant files will be extracted
        :param private_key_path:
        :param immutable_dir:
        :param mutable_dir:
        :return:
        """
        logging.info("Executing pre-run protocol...")
        # Execute the protocol with directly passed files and the instances config file
        if img and private_key_path:
            logging.info("Extracting files from image...")
            # Get the content of the immutable files from the image as ByteObjects
            immutable_files, file_names = files_from_archive(extract_archive(img, immutable_dir))

            # Check that no files have been added or removed
            assert len(immutable_files) == len(self.key_manager.get_security_param("immutable_file_list"))

            self.validate_immutable_files(
                files=immutable_files,
                immutable_file_names=file_names,
                ordered_file_list=self.key_manager.get_security_param("immutable_file_list"))
            if not self._is_first_station_on_route():
                self.verify_digital_signature()
                file_encryptor = FileEncryptor(self.key_manager.get_sym_key(self.station_id,
                                                                            private_key_path=private_key_path))
                # Decrypt all previously encrypted files
                mutable_files, mf_members, mf_dir = result_files_from_archive(extract_archive(img, mutable_dir))
                decrypted_files = file_encryptor.decrypt_files(mutable_files, binary_files=True)
                self.validate_previous_results(files=decrypted_files)
                archive = self._make_results_archive(mf_dir, mf_members, decrypted_files)
                logging.info("Adding decrypted files to image")
                # print(archive.name)
                self._update_image(img, archive, results_path="/opt")

            logging.info("Pre-run protocol success")
            return
        # Execute the protocol parsing the files from the given train and results directories
        elif self.results_dir and self.train_dir:
            self.validate_immutable_files(self.train_dir)

            if not self._is_first_station_on_route():
                self.verify_digital_signature()

                files = self._parse_files(self.results_dir)

                file_encryptor = FileEncryptor(self.key_manager.get_sym_key(self.station_id))

                # Decrypt all previously encrypted files
                file_encryptor.decrypt_files(files)
                self.validate_previous_results()

            logging.info("Pre-run protocol success")
            return
        else:
            raise ValueError("Neither train image or results and train directories are set")

    def post_run_protocol(self, img: str = None, private_key_path: str = None, mutable_dir: str = "/opt/pht_results"):
        """
        Updates the necessary values in the train_config.json and encrypts the updated files after a successful train
        execution.

        :param img: identifier of the image <repository>:<tag>
        :param private_key_path: path to the private key associated with the current station and with the corresponding
            public key registered in vault under the PID chosen by the station
        :param mutable_dir: path to the directory in which the mutable files are stored
        :return:
        """
        # execute the post run station side extracting the relevant files from the image
        if img and private_key_path:
            logging.info(f"Executing post-run protocol - target image: {img} \n")
            # Get the mutable files and tar archive structure
            mutable_files, mf_members, mf_dir = result_files_from_archive(extract_archive(img, mutable_dir))
            # Run the post run protocol
            encrypted_mutable_files = self._post_run_outside_container(mutable_files, private_key_path)
            results_archive = self._make_results_archive(mf_dir, mf_members, encrypted_mutable_files)
            results_archive.seek(0)

            # update the container with the encrypted files
            self._update_image(img, results_archive, results_path="/opt", config_path="/opt")
            logging.info(f"Successfully executed post run protocol on img: {img}")
        # execute the post run protocol running inside the docker container
        else:
            logging.info("Executing post-run protocol: \n")
            self._post_run_in_container()

    def _post_run_outside_container(self, mutable_files: List[BytesIO], private_key_path: str) -> List[BytesIO]:
        """
        Performs the post run protocol on the mutable files contained in an image. Consisting of encrypting the
        mutable files and updateing the train_config.json. The extracted mutable files are encrypted and the
        train_config.json is updated to reflect the current state of the train.
        The changed files are written to the base image and the resulting image is tagged as latest.

        :param mutable_files: list of BytesIO objects containing the mutable files for a train
        :param private_key_path: path to private key used to sign the results
        :return:
        """
        logging.info(f"prev results hash {self.key_manager.get_security_param('e_d')}", )
        # Update the hash value of the mutable files
        e_d = hash_results(result_files=mutable_files,
                           session_id=bytes.fromhex(self.key_manager.get_security_param("session_id")),
                           binary_files=True)
        self.key_manager.set_security_param("e_d", e_d.hex())
        logging.info(f"new results hash: {self.key_manager.get_security_param('e_d')}")

        # Load the local private key and sign the hash of the results files
        sk = self.key_manager.load_private_key(key_path=private_key_path)
        e_d_sig = sk.sign(e_d,
                          padding.PSS(mgf=padding.MGF1(hashes.SHA512()),
                                      salt_length=padding.PSS.MAX_LENGTH),
                          utils.Prehashed(hashes.SHA512()))
        self.key_manager.set_security_param("e_d_sig", e_d_sig.hex())

        # Update the digital signature of the train
        self.sign_digital_signature(sk)

        for file in mutable_files:
            file.seek(0)
        # Encrypt the files
        new_sym_key = self.key_manager.generate_symmetric_key()
        file_encryptor = FileEncryptor(new_sym_key)
        encrypted_results = file_encryptor.encrypt_files(mutable_files, binary_files=True)

        # Encrypt the new symmetric key with the available public keys and store them in the image
        encrypted_symmetric_keys = self.key_manager.encrypt_symmetric_key(new_sym_key)
        self.key_manager.set_security_param("encrypted_key", encrypted_symmetric_keys)
        # at the last station encrypt the symmetric key using the rsa public key of the user

        user_public_key = self.key_manager.load_public_key(
            self.key_manager.get_security_param("rsa_user_public_key"))
        user_encrypted_sym_key = self.key_manager._rsa_pk_encrypt(new_sym_key, user_public_key)
        self.key_manager.set_security_param("user_encrypted_sym_key", user_encrypted_sym_key)

        return encrypted_results

    def _update_image(self, img, results_archive: BytesIO, results_path: str, config_path: str = None):
        """
        Update the base image with the encrypted results files and the updated train_config.json and tag it as
        latest
        :param img: identifier of the image <repository>:<tag>
        :param results_archive: tar archive containing the encrypted results
        :param results_path: path to write the results to
        :param config_path: path to write the updated train_config.json
        :return:
        """
        # If a config path is given update the train config inside the container
        client = self.docker_client if self.docker_client else docker.from_env()
        base_image = img.split(":")[0] + ":" + "base"
        container = client.containers.create(base_image)

        if config_path:
            logging.info("Updating train config")
            config_archive = self._make_train_config_archive()
            config_archive.seek(0)
            # add_archive(img, config_archive, config_path)
            container.put_archive(config_path, config_archive)
        # add the updated results archive
        logging.info("Adding encrypted result files")
        container.put_archive(results_path, results_archive)
        # add_archive(img, results_archive, results_path)
        logging.info("Updating user key file ")
        user_key = self._make_user_key()
        # add user key to opt directory
        # add_archive(img, user_key, "/opt")
        container.put_archive("/opt", user_key)
        # Tag container as latest
        repo, tag = img.split(":")
        container.commit(repository=repo, tag=tag)
        container.wait()
        container.remove()

    @staticmethod
    def _make_results_archive(archive_members, file_members, updated_files):
        """
        Creates a tar archive containing the updated results

        :param archive_members: tar directory structure of the pht_results directory
        :param file_members: the members of the archive representing actual files
        :param updated_files: updated files associated with the file_members that will be written to the archive
        :return: updated tar archive to be copied into the new image.
        """
        archive_obj = BytesIO()
        tar = tarfile.open(fileobj=archive_obj, mode="w")

        # Create directory structure
        for member in archive_members:
            if member not in file_members:
                tar.addfile(member)
        # Add the updated members to a new archive
        for i, file_member in enumerate(file_members):
            file_size = updated_files[i].getbuffer().nbytes
            updated_files[i].seek(0)
            file_member.size = file_size
            file_member.mtime = time.time()
            tar.addfile(file_member, fileobj=updated_files[i])

        # Close tarfile and reset BytesIO
        tar.close()
        archive_obj.seek(0)

        return archive_obj

    def _make_train_config_archive(self) -> BytesIO:
        """
        Create in memory tar archive containing the train configuration json file
        :return:
        """
        archive_obj = BytesIO()
        tar = tarfile.open(fileobj=archive_obj, mode="w")
        data = self.key_manager.save_keyfile(binary_file=True)

        # Create TarInfo Object based on the data
        info = TarInfo(name="train_config.json")
        info.size = data.getbuffer().nbytes
        info.mtime = time.time()
        # add config data and reset the archive
        tar.addfile(info, data)
        tar.close()
        archive_obj.seek(0)
        return archive_obj

    def _make_user_key(self):
        archive_obj = BytesIO()
        tar = tarfile.open(fileobj=archive_obj, mode="w")
        # Extract user key from config and convert it to bytesio
        data = BytesIO(bytes.fromhex(self.key_manager.get_security_param("user_encrypted_sym_key")))
        info = TarInfo(name="user_sym_key.key")
        info.size = data.getbuffer().nbytes
        info.mtime = time.time()
        tar.addfile(info, data)
        tar.close()
        archive_obj.seek(0)
        return archive_obj

    def _post_run_in_container(self):
        """
        Execute the post-run protocol inside of the container

        :return:
        """

        # Update the values hash and signature of the results
        files = self._parse_files(self.results_dir)
        e_d = hash_results(files, bytes.fromhex(self.key_manager.get_security_param("session_id")))
        self.key_manager.set_security_param("e_d", e_d.hex())
        sk = self.key_manager.load_private_key(env_key="RSA_STATION_PRIVATE_KEY")
        e_d_sig = sk.sign(e_d,
                          padding.PSS(mgf=padding.MGF1(hashes.SHA512()),
                                      salt_length=padding.PSS.MAX_LENGTH),
                          utils.Prehashed(hashes.SHA512()))
        self.key_manager.set_security_param("e_d_sig", e_d_sig.hex())
        # Sign the train after execution
        self.sign_digital_signature(sk)
        # Write new values to file and encrypt files with new symmetric key
        new_sym_key = self.key_manager.generate_symmetric_key()
        file_encryptor = FileEncryptor(new_sym_key)

        # Encrypt the files
        file_encryptor.encrypt_files(files)
        self.key_manager.set_security_param("encrypted_key", self.key_manager.encrypt_symmetric_key(new_sym_key))
        # at the last station encrypt the symmetric key using the rsa public key of the user
        user_public_key = self.key_manager.load_public_key(
            self.key_manager.get_security_param("rsa_user_public_key"))
        user_encrypted_sym_key = self.key_manager._rsa_pk_encrypt(new_sym_key, user_public_key)
        self.key_manager.set_security_param("user_encrypted_sym_key", user_encrypted_sym_key)

        self.key_manager.save_keyfile()
        logging.info("Post-protocol success")

    def validate_immutable_files(self, train_dir: str = None, files: list = None, ordered_file_list: List[str] = None,
                                 immutable_file_names: List[str] = None):
        """
        Checks if the hash of the immutable files is the same as the one stored at the creation of the train

        :raises ValidationError: when the files to be executed do not match the agreed upon files

        :return:
        """
        # check the signature of the stored hash value using ec signature verifying that it is created by the user
        user_pk = self.key_manager.load_public_key(self.key_manager.get_security_param("rsa_user_public_key"))
        e_h = bytes.fromhex(self.key_manager.get_security_param("e_h"))
        e_h_sig = bytes.fromhex(self.key_manager.get_security_param("e_h_sig"))
        # now check before the run that no immutable files have changed, based on stored hash
        if train_dir:
            immutable_files = self._parse_files(train_dir)
            immutable_files = [str(file) for file in immutable_files if "train_config.json" not in str(file)]

            current_hash = hash_immutable_files(immutable_files, str(self.key_manager.get_security_param("user_id")),
                                                bytes.fromhex(self.key_manager.get_security_param("session_id")),
                                                )
        elif files:
            current_hash = hash_immutable_files(files,
                                                str(self.key_manager.get_security_param("user_id")),
                                                bytes.fromhex(self.key_manager.get_security_param("session_id")),
                                                binary_files=True,
                                                ordered_file_list=ordered_file_list,
                                                immutable_file_names=immutable_file_names
                                                )
        logging.info(f"e_h: {e_h}")
        logging.info(f"file hash: {current_hash}")
        if e_h != current_hash:
            raise ValidationError("Immutable Files have changed")
        # Verify that the hash value corresponds with the signature
        user_pk.verify(e_h_sig,
                       current_hash,
                       padding.PSS(mgf=padding.MGF1(hashes.SHA512()),
                                   salt_length=padding.PSS.MAX_LENGTH),
                       utils.Prehashed(hashes.SHA512()))

    def validate_previous_results(self, files: List[BinaryIO] = None):
        """
        Verify that the results from the execution of the previous station did not change, by hashing the stored results
        from the previous station and comparing it with the decrypted stored hash from the previous station
        """
        # verify the hash of the results of the previous station
        prev_results_hash = self.key_manager.get_security_param("e_d")
        results_sig = self.key_manager.get_security_param("e_d_sig")
        # Load the public key of the station
        ds = self.key_manager.get_security_param("digital_signature")
        # Get public key of the previous station
        station_public_key = self.key_manager.get_security_param("rsa_public_keys")[ds[-1]["station"]]
        station_public_key = self.key_manager.load_public_key(station_public_key)
        if files:
            results_hash = hash_results(files,
                                        session_id=bytes.fromhex(self.key_manager.get_security_param("session_id")),
                                        binary_files=True)
        else:
            files = self._parse_files(self.results_dir)
            results_hash = hash_results(files, bytes.fromhex(self.key_manager.get_security_param("session_id")))
        station_public_key.verify(bytes.fromhex(results_sig),
                                  results_hash,
                                  padding.PSS(mgf=padding.MGF1(hashes.SHA512()),
                                              salt_length=padding.PSS.MAX_LENGTH),
                                  utils.Prehashed(hashes.SHA512()))
        # Compare with the files currently present in the train
        if results_hash != bytes.fromhex(prev_results_hash):
            raise ValidationError("The previously hashed results do not match the stored ones")

    def sign_digital_signature(self, sk: RSAPrivateKey):
        """
        Update the digital signature of the train after successful execution of the train.
        If there is no previous signature present creates a signature based on the session id, otherwise the
        signature of the previous station is loaded, signed using the current stations private key and appended to
        the list of signatures stored in the train.

        :param sk: private key of the currently running station
        """
        ds = self.key_manager.get_security_param("digital_signature")
        hasher = hashes.Hash(hashes.SHA512(), default_backend())
        if ds is None:
            hasher.update(bytes.fromhex(self.key_manager.get_security_param("session_id")))
            digest = hasher.finalize()
            sig = sk.sign(digest,
                          padding.PSS(mgf=padding.MGF1(hashes.SHA512()),
                                      salt_length=padding.PSS.MAX_LENGTH),
                          utils.Prehashed(hashes.SHA512())
                          )
            ds = [{"station": self.station_id, "sig": (sig.hex(), digest.hex())}]
            self.key_manager.set_security_param("digital_signature", ds)
        else:
            hasher.update(bytes.fromhex(ds[-1]["sig"][0]))
            digest = hasher.finalize()
            sig = sk.sign(digest,
                          padding.PSS(mgf=padding.MGF1(hashes.SHA512()),
                                      salt_length=padding.PSS.MAX_LENGTH),
                          utils.Prehashed(hashes.SHA512())
                          )
            ds.append({"station": self.station_id, "sig": (sig.hex(), digest.hex())})
            self.key_manager.set_security_param("digital_signature", ds)

    def verify_digital_signature(self):
        """
        Verifies the digital signature of the train by iterating over the list of signatures and verifying each one
        using the correct public key stored in the train configuration json

        :raise: InvalidSignatureError if any of the signed values can not be validated using the provided public keys

        """
        ds = self.key_manager.get_security_param("digital_signature")
        for sig in ds:
            public_key = self.key_manager.load_public_key(
                self.key_manager.get_security_param("rsa_public_keys")[sig["station"]])
            public_key.verify(bytes.fromhex(sig["sig"][0]),
                              bytes.fromhex(sig["sig"][1]),
                              padding.PSS(mgf=padding.MGF1(hashes.SHA512()),
                                          salt_length=padding.PSS.MAX_LENGTH),
                              utils.Prehashed(hashes.SHA512())
                              )

    def _is_first_station_on_route(self) -> bool:
        """
        Returns true if current station is the first station on the route
        :return:
        """
        # Check if there are previous results if not station is first station on route
        return self.key_manager.get_security_param("e_d") is None

    @staticmethod
    def _parse_files(target_dir):
        """
        Parses the exported files from a container and sorts them into relevant categories
        :param target_dir: directory in which to find all files
        :return: Tuple consisting of lists of paths for the different file types
        """
        files = list()
        logging.info("Detecting files...")
        for (dir_path, dir_names, file_names) in os.walk(target_dir):
            files += [os.path.join(dir_path, file) for file in file_names]
        logging.info(f"Found {len(files)} Files")
        return files




from io import BytesIO
from typing import Union

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import padding
from cryptography.hazmat.primitives import hashes
from cryptography.fernet import Fernet
import os
import json


class KeyManager:
    """
    Class that creates, stores and if necessary updates all relevant keys for symmetric and asymmetric encryption
    """

    def __init__(self, train_config: Union[str, dict]):
        """
        Initialize a KeyManager instance that handles security relevant values based on a train configuration
        
        :param train_config: either a path to a json file storing the configuration values or a dictionary with these values 
        """
        self.config_path = None
        if type(train_config) == dict:
            self.config = train_config
        else:
            self.config_path = train_config
            with open(train_config, "r") as config_file:
                self.config = json.load(config_file)

    def save_keyfile(self, path=None, binary_file=False):
        """
        Store the updated config file as a json at the same location

        :return:
        :rtype:
        """
        if binary_file:
            return BytesIO(json.dumps(self.config).encode("utf-8"))

        if not self.config_path and path:
            raise ValueError("To save the path either a path needs to given as argument or set as an instance variable")
        if path:
            self.config_path = path
        with open(self.config_path, "w") as config_file:
            json.dump(self.config, config_file, indent=2)

    def get_security_param(self, param: str):
        """
        Returns a parameter from the associated keyfile
        :param param:
        :return: value of the specified parameter
        """
        return self.config[param]

    def set_security_param(self, param: str, value):
        """
        Updates a parameter in the keyfile with the given value
        :param param: the parameter to update
        :param value: new value for param
        :return:
        """
        self.config[param] = value

    @staticmethod
    def generate_symmetric_key():
        """
        Create a symmetric fernet key for encrypting sensitive files
        :return:
        """
        return Fernet.generate_key()

    def get_sym_key(self, station_id: str, private_key_path: str = None):
        """
        Decrypts the symmetric key using a stored private key
        :arg station_id: station identifier used to load the correct public key
        :return: symmetric fernet key used to encrypt and decrypt files
        """
        if private_key_path:
            private_key = self.load_private_key(key_path=private_key_path)
        else:
            private_key = self.load_private_key(env_key="RSA_STATION_PRIVATE_KEY")
        encrypted_sym_key = self.get_security_param("encrypted_key")[station_id]
        symmetric_key = private_key.decrypt(bytes.fromhex(encrypted_sym_key),
                                            padding.OAEP(
                                                mgf=padding.MGF1(algorithm=hashes.SHA512()),
                                                algorithm=hashes.SHA512(),
                                                label=None
                                            )
                                            )
        return symmetric_key

    def generate_encrypted_keys(self, symmetric_key: bytes):
        """
        Generates a dictionary containing the symmetric key used to encrypt files, encrypted with the public keys of all
        stations on the route
        :param symmetric_key: byte object containing the symmetric key used to encrypt the mutable files
        :return: Dictionary consisting of  key = Station Id, value = Symmetric key encrypted with public key of station
        """
        enc_keys = {}
        for station, pk in self.config["rsa_public_keys"]:
            enc_keys[station] = self.encrypt_symmetric_key(symmetric_key, pk)
        return enc_keys

    def encrypt_symmetric_key(self, sym_key: bytes) -> dict:
        """
        Encrypt the symmetric key with all public keys provided in the train configuration file

        :param sym_key: byte object containing the the symmetric key used to encrypt the mutable files
        :return: dictionary containing the symmetric key encrypted with all available public keys, keys are the station
        ids and values are the symmetric key encrypted with the RSA public key associated with the station id

        :rtype: dict
        """
        encrypted_keys = {}
        for id, key in self.config["rsa_public_keys"].items():
            public_key = self.load_public_key(key)
            encrypted_key = public_key.encrypt(sym_key,
                                               padding.OAEP(mgf=padding.MGF1(algorithm=hashes.SHA512()),
                                                            algorithm=hashes.SHA512(),
                                                            label=None)
                                               )
            encrypted_keys[id] = encrypted_key.hex()
        return encrypted_keys

    def _rsa_pk_encrypt(self, val, public_key):
        encrypted = public_key.encrypt(val,
                                       padding.OAEP(mgf=padding.MGF1(algorithm=hashes.SHA512()),
                                                    algorithm=hashes.SHA512(),
                                                    label=None)
                                       )
        return encrypted.hex()

    @staticmethod
    def load_private_key(env_key: str = None, key_path: str = None):
        """
        Loads the private key from the path provided provided in the environment variables of the currently
        running image
        :param key_path: path to a file containing the private key
        :param env_key: environment variable containing a hex string representing the station private key
        :return: a private key object either rsa or ec
        """

        if env_key and key_path:
            raise ValueError(f"Multiple private Key locations specified: \n {env_key} \n {key_path}")
        # Load key from hex string stored in environment variable
        if env_key:
            private_key = serialization.load_pem_private_key(bytes.fromhex(os.getenv(env_key)),
                                                             password=None,
                                                             backend=default_backend())
        # Load key from file
        elif key_path:
            with open(key_path, "rb") as sk_f:
                private_key = serialization.load_pem_private_key(sk_f.read(),
                                                                 password=None,
                                                                 backend=default_backend()
                                                                 )
        else:
            raise ValueError("No environment variable or file containing a private key specified")

        return private_key

    @staticmethod
    def load_public_key(key: str):
        """
        Loads a public key
        :param key: string representation of a public key
        :return: public key object for asymmetric encryption
        """
        public_key = serialization.load_pem_public_key(bytes.fromhex(key),
                                                       backend=default_backend())
        return public_key

